/*
 * SuSE Checksum by Marc Heuse <marc@suse.de>
 *
 */

#define VERSION "v1.9"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#define LENGTH   4100

#ifndef major  /* hmmm defined only in some sys/types.h - weird. Thanks afx */
#define major(dev) (((dev) >> 8) & 0xff)
#define minor(dev) ((dev) & 0xff)
#endif

int absolute = 0, debug = 0, ignore = 0, some_partitions_only = 0, setid = 0, length = 0;
int isdir = 0, checkmode = 0, force = 0, partition_no = 0, missing = 0;
char sumdata[LENGTH+900];
unsigned long partitions[25];
FILE *output;

void create_checksum(char *filename, char *buf) {
    struct stat st;
    unsigned char hash[33] = "HASH_ERROR", binhash[16];
    char type = '\0';
    FILE *fp;
    unsigned int size = 0;

    if (debug) fprintf(stderr, "create_checksum(%s)\n", filename);

    if (lstat(filename, &st) < 0) {
	perror(filename);
	snprintf(buf, 4000, "%s %d NO_PERMISSIONS-OR-FILE_NOT_FOUND", filename, strlen(filename));
	return;
    }

    if (S_ISDIR(st.st_mode)) {
        type = 'd';
        isdir = 1;
    }
    if (S_ISREG(st.st_mode)) {
        type = 'f';
        size = st.st_size;
    }
    if (S_ISLNK(st.st_mode))
        type = 'l';
    if (S_ISCHR(st.st_mode))
        type = 'c';
    if (S_ISBLK(st.st_mode))
        type = 'b';
    if (S_ISFIFO(st.st_mode))
        type = 'p';
    if (S_ISSOCK(st.st_mode))
        type = 's';
    if (type == '\0')
        type = '?';

    if (some_partitions_only) {
        int ii = 0;
        do {
            ii++;
        } while ((st.st_dev != partitions[ii-1]) && (ii < partition_no));
        if (st.st_dev != partitions[ii-1]) {
            strcpy(buf, "");
            return;
        }
    }
    if (setid) {
        if ((! S_ISREG(st.st_mode)) || ((st.st_mode & (S_ISUID + S_ISGID)) == 0)) {
            strcpy(buf, "");
            return;
        }
    }


    if ((force > 0 || (type == 'f' && size > 0)) && access(filename, R_OK) == 0) {
	if ((fp = fopen(filename, "r")) == NULL) {
	    perror(filename);
	    strcpy(hash, "HASH_FAILED");
	}

	md5_stream(fp, binhash, length);
	sprintf(hash, "%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x%.2x",
		binhash[0], binhash[1],	binhash[2], binhash[3], binhash[4], binhash[5],
		binhash[6], binhash[7], binhash[8], binhash[9], binhash[10],
		binhash[11], binhash[12], binhash[13], binhash[14], binhash[15]);
	fclose(fp);
    } else
        strcpy(hash, "NO_HASH");
    
    /* filename uid gid permissions type size hash */
    if (type != 'b' && type != 'c')
        snprintf(buf, 4090, "%s %d %d %o %c %lu %d %s", filename, st.st_uid,
            st.st_gid, st.st_mode & 07777, type, size, strlen(filename), hash);
    else
        snprintf(buf, 4090, "%s %d %d %o %c %u,%u 0 %d %s", filename, st.st_uid, 
            st.st_gid, st.st_mode & 07777, type, (unsigned char) major(st.st_rdev),
            (unsigned char) minor(st.st_rdev), strlen(filename), hash);
    return;
}

int do_ignore(char *filename) {
    char *ptr;

    if (ignore == 0)
        return 0;
    if (((char **)ptr = rindex(filename, '/')) == NULL)
        ptr = filename;
    else
        ptr++;
    
    if (strcmp("tmp", ptr) && strcmp("proc", ptr))
        return 0;
    return 1;
}

void go_recursive(char *filename, int recursive_level) {
    DIR *dir;
    struct dirent *dir_entry;
    char current_dir[4097], chdir_dir[4097];

    isdir = 0;
    getcwd(current_dir, 4096);
    current_dir[4096] = '\0';
    if (chdir(filename)) {
        perror(filename);
        return;
    }
    if ((dir = opendir (".")) == NULL) {
        perror(filename);
        return;
    }
    getcwd(chdir_dir, 4096);
    chdir_dir[4096] = '\0';
    while ((dir_entry = readdir(dir)) != NULL) {
        sumdata[0] = '\0';
        isdir = 0;
	if (strcmp(dir_entry->d_name, ".") && strcmp(dir_entry->d_name, "..")) {
            create_checksum(dir_entry->d_name, sumdata);
	    if (strlen(sumdata) > 0) {
	        if (absolute && sumdata[0] != '/')
	            if (strlen(chdir_dir) > 1)
                        fprintf(output,"%s/%s\n", chdir_dir, sumdata);
                    else
                        fprintf(output,"%s%s\n", chdir_dir, sumdata);
                else
            	    fprintf(output,"%s\n", sumdata);
            }
            if (isdir && recursive_level && do_ignore(dir_entry->d_name) == 0)
                go_recursive(dir_entry->d_name, recursive_level);
	}
    }
    closedir(dir);
    if(chdir(current_dir) != 0) {
        perror("Can not change to old directory, aborting");
        exit(-1);
    }
    return;
}

void make_absolute(char *frel, char *fabs) {
    char path1[LENGTH] = "", path2[LENGTH] = "", tmp[LENGTH+100] = "";
    char *p;
    
    if (frel[0] == '/') {
        strcpy(fabs, frel);
    } else {
        strcpy(tmp, frel);
        getcwd(path1, LENGTH - 1);
        path1[sizeof(path1)-1] = '\0';
        if (((char **)p = rindex(tmp, '/')) != NULL) {
	    *p = '\0';
	    p++;
	    if (chdir(tmp) < 0) {
	        fprintf(stderr, "Can not determine absolute filename of %s.\n", frel);
	        strcpy(fabs, frel);
	        return;
	    }
	    getcwd(path2, LENGTH - 1);
	    path2[sizeof(path2)-1] = '\0';
	    if (chdir(path1) < 0) {
	        perror("Can not change to old directory, aborting");
	        exit(-1);
	    }
	    strcpy(fabs, path2);
	    strcat(fabs, "/");
	    strncat(fabs, p, LENGTH*2 - strlen(fabs) - 1);
        } else {
            strcpy(fabs, path1);
            strcat(fabs, "/");
            strncat(fabs, frel, LENGTH*2 - strlen(fabs) - 1);
        }
    }
    return;
}

void check4new(char *newfile) {

/* XXX */

}

int main(int argc, char *argv[]) {
    int res, recursive = 0, found = 0, fd;
    struct stat st;
    char checkfile[2*LENGTH+300];
    char file[LENGTH] = "", infile[LENGTH] = "", path[LENGTH] = "", upfile[LENGTH] = "";
    char savedsumdata[2*LENGTH+300];
    char *p, *tmpchar;
    FILE *f;

    if ((argc < 2) || (strcmp(argv[1], "-h") == 0)) {
        fprintf(stderr, "SuSE Checksum %s by Marc Heuse <marc@suse.de>  http://www.suse.de/~marc\n", VERSION);
	fprintf(stderr, "Syntax: %s [-adirsF] [-f file] [-l length] [-m path] file(s)\n", argv[0]);
 	fprintf(stderr, "Syntax: %s [-adirsF] [-M] [-l length] [-m path] -u file\n", argv[0]);
	fprintf(stderr, "Syntax: %s [-l length] [-p path] -c file\n", argv[0]);
	fprintf(stderr, "Options for CREATING checksums:\n");
	fprintf(stderr, "\t-l length only process the first length bytes of the file (L <= 4096)\n");
	fprintf(stderr, "\t-f file   read filenames from file to create checksums for\n");
	fprintf(stderr, "\t-o file   output file for the checksum/differences (defaults to stdout)\n");
	fprintf(stderr, "\tfile(s)   file(s) to create a checksum for\n");
	fprintf(stderr, "Options for UPDATING checksums:\n");
	fprintf(stderr, "\t-M        report missing and new files (and adding them)\n");
	fprintf(stderr, "\t-l length only process the first length bytes of the file (L <= 4096)\n");
	fprintf(stderr, "\t-o file   output file for the checksum/differences (defaults to stdout)\n");
	fprintf(stderr, "\t-u file   file with checksumdata to check & update\n");
	fprintf(stderr, "Options for VERIFYING checksums: (note that -F is always set here)\n");
	fprintf(stderr, "\t-l length only process the first length bytes of the file (L <= 4096)\n");
	fprintf(stderr, "\t-p path   prepend path to saved filenames to check\n");
	fprintf(stderr, "\t-M        report missing and new files\n");
	fprintf(stderr, "\t-o file   output file for the checksum/differences (defaults to stdout)\n");
	fprintf(stderr, "\t-c file   file with checksumdata to check against\n");
	fprintf(stderr, "Options for CREATING and add-on options for -M:\n");
	fprintf(stderr, "\t-a        print absolute pathnames\n");
	fprintf(stderr, "\t-i        ignore ALL \"tmp\" and \"proc\" directories and below\n");
	fprintf(stderr, "\t-F        force to make md5 checksums on all targets (devices!)\n");
	fprintf(stderr, "\t-d        if file is a directory, process also it's contents\n");
	fprintf(stderr, "\t-m path   only process files on this mounted partition(s) (no sub part.)\n");
	fprintf(stderr, "\t-r        if file is a directory, process it recursive\n");
	fprintf(stderr, "\t-s        checksum only set-uid and set-gid files\n");
	fprintf(stderr, "Information/Errors are printed to stderr, data and results are printed to\n");
	fprintf(stderr, "stdout. Exits with -1 on error, 0 if OK and 1 if -c finds differences.\n");
	exit(-1);    }
    output = stdout;

    while(1) {
	res = getopt(argc, argv, "ac:def:il:m:o:p:rsu:DFM");
	if (res < 0) break;
	switch (res) {
	    case 'a' : absolute = 1;
	    	  break;
	    case 'c' : strncpy(file, optarg, sizeof(file)-1);
	          file[sizeof(file)-1] = '\0';   
	    	  break;
	    case 'd' : recursive += 1;
	    	  break;
	    case 'e' :
	    	  fprintf(stderr,"The option -e is obsolete. Use -m path.\n");
	    	  exit(-1);
	    case 'm' :
	    	  tmpchar = malloc(4097);
	    	  some_partitions_only = 1;
	    	  strncpy(tmpchar, optarg, 4090);
	    	  strcat(tmpchar, "/.");
	          partition_no++;
	          if (partition_no > 25) {
		      fprintf(stderr,"Only up to 25 partitions are supported.\n");
		      exit(-1);
	          }
	          if (stat(optarg, &st) < 0) {
		      fprintf(stderr,"Can not find mounted partition %s\n",optarg);
		      exit(-1);
	          }
		  partitions[partition_no-1] = st.st_dev;
	          free(tmpchar);
	          break;
	    case 'f' : strncpy(infile, optarg, sizeof(infile)-1);
	          infile[sizeof(infile)-1] = '\0';
	          break;
	    case 'i' : ignore = 1;
	          break;
	    case 'l' : length = atoi(optarg);
	          break;
	    case 'o' : unlink(optarg);
	          if ((fd=open(optarg,O_WRONLY|O_CREAT|O_EXCL)) < 0) {
	              fprintf(stderr,"Could not create %s\n",optarg);
	              exit(-1);
	          }
	          if((output=fdopen(fd,"w")) == NULL) {
	              fprintf(stderr,"Could not create %s\n",optarg);
	              exit(-1);
	          }
	          break;
	    case 'p' : strncpy(path, optarg, LENGTH - 1);
	          path[LENGTH - 1] = '\0';
	          if (strlen(path) > 0)
	              if (path[strlen(path)-1] == '/')
	                  path[strlen(path)-1] = '\0';
	          break;
	    case 'r' : recursive+=2;
	    	  break;
	    case 's' : setid = 1;
	          break;
	    case 'u' : strncpy(upfile, optarg, LENGTH - 1);
	          upfile[LENGTH - 1] = '\0';
	          if (strlen(upfile) > 0)
	              if (path[strlen(upfile)-1] == '/')
	                  upfile[strlen(upfile)-1] = '\0';
	          break;
	    case 'D' : debug++;
	          break;
	    case 'F' : force = 1;
	          break;
	    case 'M' : missing = 1;
	          break;
	    default: exit(-1);
	}
    }
    res = optind;
    if (length < 0 || length > 4096) {
	fprintf(stderr, "Length must be greater than 0 and less/equal to 4096: %d\n", length);
	exit(-1);
    }
    if (recursive > 2) {
	fprintf(stderr, "You may only set either -d or -r.\n");
	exit(-1);
    }
    fd=((strlen(infile)>0 || res < argc) ? 1 : 0) + (strlen(file)>0 ? 1 : 0) + (strlen(upfile)>0 ? 1 : 0);
    if (fd == 0) {
	fprintf(stderr,"I don't know what to do. See \"%s -h\" and try again.\n",argv[0]);
	exit(-1);
    }
    if (fd > 1) {
	fprintf(stderr,"You can only choose one mode: -u for update, -c for checking/verifying, or -f
and/or filenames on the commandline for creating checksums.\n");
	exit(-1);
    }
    if ((absolute || some_partitions_only || ignore || missing || recursive || res < argc) && strlen(file) > 0) {
	fprintf(stderr, "You may only choose either validating a checksum file with -c OR a set of files and other options.\n");
	exit(-1); 
    }
    
    if (strlen(file) > 0 || strlen(upfile) > 0) { // CHECKING/VERIFYING function
        char tmpfile[LENGTH] = "";
        char *dof;
        FILE *tf = NULL;
        unsigned int flength;
        checkmode = 1; /*force = 1;*/
        if (strlen(file) > 0)
            dof = file;
        else {
            char *tmpbak = malloc(4111);
            int ii = 0;
            tmpchar = malloc(4111);
            dof = upfile;
	    strncpy(tmpfile, upfile, LENGTH-10);
	    tmpfile[LENGTH-10] = '\0';
	    strcat(tmpfile, ".XXXXXX");
            if ((fd = mkstemp(tmpfile)) < 0) {
	        fprintf(stderr,"Could not create temporary update file %s\n",tmpfile);
	        exit(-1);
            }
            tf=fdopen(fd,"w");
            sprintf(tmpchar, "##//## called as \"");
            for (ii = 0; ii < argc; ii++) {
                sprintf(tmpbak, "%s%s ",tmpchar,argv[ii]);
                strcpy(tmpchar, tmpbak);
            }
            tmpchar[strlen(tmpchar)-1] = '"';
            fprintf(tf,"%s\n",tmpchar);
            fprintf(tf,"##//## database syntax: FILENAME UID GID MODE TYPE SIZE FILENAME_LENGTH CHECKSUM\n");
            free(tmpchar);
            free(tmpbak);
        }
        if ((f = fopen(dof, "r")) == NULL) {
            perror(file);
            if (strlen(tmpfile)>0)
                unlink(tmpfile);
            exit(-1);
        }
	while(! feof(f)) {
	    fgets(savedsumdata, sizeof(savedsumdata), f);
	    if ((strncmp(savedsumdata, "##//##", 6) == 0) || (strlen(savedsumdata) < 2))
	        continue;
	    if (strlen(savedsumdata) > 0) {
	        if (savedsumdata[strlen(savedsumdata)-1] != '\n' && ! feof(f)) {
		    fprintf(stderr, "(1) Error in this line of sumfile %s: %s\n", file, savedsumdata);
		    exit(-1);
	        } else
	            savedsumdata[strlen(savedsumdata)-1] = '\0';
	        strncpy(checkfile, savedsumdata, sizeof(checkfile)-1);
	        checkfile[sizeof(checkfile)-1] = '\0';
	        if (((char **)p = rindex(checkfile, ' ')) == NULL) {
	            fprintf(stderr, "(2) Error in this line of sumfile %s: %s\n", file, savedsumdata);
	            exit(-1);
	        }
	        *p = '\0';
	        if (((char **)p = rindex(checkfile, ' ')) == NULL) {
		    fprintf(stderr, "(3) Error in this line of sumfile %s: %s\n", file, savedsumdata);
		    exit(-1);
	        }
	        *p = '\0';
	        p++;
	        flength = atoi(p);
	        if (flength < 1 || flength > 4096) {
		    fprintf(stderr, "(3) Error in this line of sumfile %s: %s\n", file, savedsumdata);
		    exit(-1);
	        }
	        checkfile[flength] = '\0';

	        if (strlen(path) > 0) {
	            strcpy(infile, checkfile);
	            strcpy(checkfile, path);
	            if (infile[0] != '/')
	                strcat(checkfile, "/");
	            strcat(checkfile, infile);
	        }
	        create_checksum(checkfile, sumdata);
	        p = sumdata;
	        if (strlen(path) > 0) {
	            p += strlen(path);
	            if (*p != '/')
	    	    p--;
	        }
	        if (strcmp(p, savedsumdata)) {
	            fprintf(output,"OLD: %s\nNEW: %s\n\n", savedsumdata, sumdata);
	            found++;
	        }
	        if (tf != NULL)
	            fprintf(tf,"%s\n",p);
		if (missing)
	            check4new(checkfile);
	    }
	    savedsumdata[0] = '\0';
	}
	fclose(f);
	if (missing)
	    check4new("");
	if (tf != NULL) {
	    fclose(tf);
	    if (rename(tmpfile,upfile) < 0)
		fprintf(stderr,"Could not rename temporary file %s to %s, I will keep both\n",tmpfile,upfile);
	}
	fprintf(stderr, "Found %d different checksums.\n", found);
	if (found > 0)
	    exit(1);
    } else {                       // CREATING function
        char *tmpbak = malloc(4111);
        int ii = 0;
        tmpchar = malloc(4111);
        checkmode = 0;

        sprintf(tmpchar, "##//## called as \"");
        for (ii = 0; ii < argc; ii++) {
            sprintf(tmpbak, "%s%s ",tmpchar,argv[ii]);
            strcpy(tmpchar, tmpbak);
        }
        tmpchar[strlen(tmpchar)-1] = '"';
        fprintf(output,"%s\n",tmpchar);
        fprintf(output,"##//## database syntax: FILENAME UID GID MODE TYPE SIZE FILENAME_LENGTH CHECKSUM\n");
        free(tmpchar);
        free(tmpbak);

	while(res < argc) {
	    if (absolute)
	        make_absolute(argv[res], checkfile);
	    else
	        strncpy(checkfile, argv[res], sizeof(checkfile)-1);
	        checkfile[sizeof(checkfile)-1] = '\0';
	    create_checksum(checkfile, sumdata);
	    if (strlen(sumdata) > 0)
	        fprintf(output,"%s\n", sumdata);
	    if (isdir && recursive && do_ignore(argv[res]) == 0)
		go_recursive(argv[res], recursive - 1);
	    res++;
	}
	if (strlen(infile) > 0) {
            if ((f = fopen(infile, "r")) == NULL) {
                perror(infile);
                exit(-1);
            }
            checkfile[0] = '\0';
            while(! feof(f)) {
                fgets(checkfile, sizeof(checkfile), f);
                if (strlen(checkfile) > 0) {
                    if (checkfile[strlen(checkfile)-1] != '\n' && ! feof(f)) {
                        fprintf(stderr, "Error in this line of infile %s: %s\n", infile, checkfile);
                        exit(-1);
                    } else
                        checkfile[strlen(checkfile)-1] = '\0';
                    create_checksum(checkfile, sumdata);
		    if (strlen(sumdata) > 0)
                        fprintf(output,"%s\n", sumdata);
                    if (isdir && recursive)
                        go_recursive(checkfile, recursive - 1);
                }
                checkfile[0] = '\0';
            }
            fclose(f);
	}
    }
    return 0;
}